/*
 * Copyright © 2008 Dennis Kasprzyk
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of
 * Dennis Kasprzyk not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.
 * Dennis Kasprzyk makes no representations about the suitability of this
 * software for any purpose. It is provided "as is" without express or
 * implied warranty.
 *
 * DENNIS KASPRZYK DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
 * NO EVENT SHALL DENNIS KASPRZYK BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: Dennis Kasprzyk <onestone@compiz-fusion.org>
 */

#ifndef _COMPPLUGINCLASSHANDLER_H
#define _COMPPLUGINCLASSHANDLER_H

#include <typeinfo>
#include <boost/preprocessor/cat.hpp>

#include <compiz.h>
#include <core/screen.h>
#include <core/pluginclasses.h>

extern unsigned int pluginClassHandlerIndex;

template<class Tp, class Tb, int ABI = 0>
class PluginClassHandler {
    public:
	PluginClassHandler (Tb *);
	~PluginClassHandler ();

	void setFailed () { mFailed = true; };
	bool loadFailed () { return mFailed; };

	Tb * get () { return mBase; };
	static Tp * get (Tb *);

    private:
	static CompString keyName ()
	{
	    return compPrintf ("%s_index_%lu", typeid (Tp).name (), ABI);
	}

    private:
	bool mFailed;
	Tb   *mBase;
};

template<class Tp, class Tb, int ABI>
PluginClassHandler<Tp,Tb,ABI>::PluginClassHandler (Tb *base) :
    mFailed (false),
    mBase (base)
{
    /* Store our index reference count as a value in screen */
    CompString  indexKeyName = compPrintf ("%s_index_%lu_structure", typeid (Tp).name (), ABI);
    CompPrivate pIndex;
    PluginClassIndex *index;
    
    if (screen->hasValue (indexKeyName))
    {
	pIndex = screen->getValue (indexKeyName);
	index = (PluginClassIndex *) pIndex.ptr;
    }
    else
    {
	index = new PluginClassIndex;
	pIndex.ptr = (void *) index;
	screen->storeValue (indexKeyName, pIndex);
    }
    
    if (index && index->failed)
    {
	mFailed = true;
    }
    else if (index)
    {
	if (!index->initiated)
	{
	    index->index = Tb::allocPluginClassIndex ();
	    if (index->index != (unsigned)~0)
	    {
		index->initiated = true;
		index->failed    = false;
		index->pcIndex = pluginClassHandlerIndex;

		CompPrivate p;
		p.uval = index->index;

		if (!screen->hasValue (keyName ()))
		{
		    screen->storeValue (keyName (), p);
		    pluginClassHandlerIndex++;
		}
		else
		{
		    compLogMessage ("core", CompLogLevelFatal,
			"Private index value \"%s\" already stored in screen.",
			keyName ().c_str ());
		}
	    }
	    else
	    {
		index->failed = true;
		index->initiated = false;
		index->pcFailed = true;
		index->pcIndex = pluginClassHandlerIndex;
		mFailed = true;
	    }
	}

	if (!index->failed)
	{
	    mBase->pluginClasses[index->index] = static_cast<Tp *> (this);
	}
    }
    else
    {
	compLogMessage ("core", CompLogLevelFatal,
	    "Index allocation for \"%s\" failed. Possibly out of memory?",
	    keyName ().c_str ());
    }
}

template<class Tp, class Tb, int ABI>
PluginClassHandler<Tp,Tb,ABI>::~PluginClassHandler ()
{
    CompString       indexKeyName = compPrintf ("%s_index_%lu_structure", typeid (Tp).name (), ABI);
    CompPrivate      pIndex = screen->getValue (indexKeyName);
    PluginClassIndex *index = (PluginClassIndex *) pIndex.ptr;

    if (index && !index->pcFailed)
    {
	/* Store our index reference count as a value in screen */
	CompString  refCountKeyName = compPrintf ("%s_index_%lu_refCnt", typeid (Tp).name (), ABI);
	CompPrivate pIndexRefCount;
	
	if (screen->hasValue (refCountKeyName))
	{
	    pIndexRefCount = screen->getValue (refCountKeyName);
	    screen->eraseValue (refCountKeyName);
	}
	else
	{
	    pIndexRefCount.uval = 0;
	}
	
	pIndexRefCount.uval--;
	screen->storeValue (refCountKeyName, pIndexRefCount);

	if (pIndexRefCount.uval == 0)
	{
	    Tb::freePluginClassIndex (index->index);
	    index->initiated = false;
	    index->failed = false;
	    index->pcIndex = pluginClassHandlerIndex;
	    screen->eraseValue (keyName ());
	    delete index;
	    screen->eraseValue (indexKeyName);
	    pluginClassHandlerIndex++;
	}
    }
}

template<class Tp, class Tb, int ABI>
Tp *
PluginClassHandler<Tp,Tb,ABI>::get (Tb *base)
{
    CompString       indexKeyName = compPrintf ("%s_index_%lu_structure", typeid (Tp).name (), ABI);
    CompPrivate      pIndex = screen->getValue (indexKeyName);
    PluginClassIndex *index = (PluginClassIndex *) pIndex.ptr;

    if (index && index->initiated && pluginClassHandlerIndex == index->pcIndex)
    {
	return static_cast<Tp *>
	    (base->pluginClasses[index->index]);
    }
    if (!index || (index->failed && pluginClassHandlerIndex == index->pcIndex))
	return NULL;

    if (index && screen->hasValue (keyName ()))
    {
	index->index     = screen->getValue (keyName ()).uval;
	index->initiated = true;
	index->failed    = false;
	index->pcIndex = pluginClassHandlerIndex;
	return static_cast<Tp *> (base->pluginClasses[index->index]);
    }
    else if (index)
    {
	index->initiated = false;
	index->failed    = true;
	index->pcIndex = pluginClassHandlerIndex;
	return NULL;
    }
    else
    {
	return NULL;
    }
}

#endif
